#!/usr/bin/env python
#
# multiclass_metrics documentation build configuration file, created by sphinx-quickstart.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))

import multiclass_metrics

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    #    'sphinx.ext.viewcode',  # create HTML file of source code and link to it
    "sphinx.ext.linkcode",  # link to github, see linkcode_resolve() below
    "recommonmark",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
# source_suffix = '.rst'
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# The master toctree document.
master_doc = "contents"

# General information about the project.
project = "multiclass-metrics"
copyright = '2024, <a href="https://www.maximzaslavsky.com">Maxim Zaslavsky</a>'
author = "Maxim Zaslavsky"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = multiclass_metrics.__version__
# The full version, including alpha/beta/rc tags.
release = multiclass_metrics.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "modules.rst"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Resolve function for the linkcode extension.
# From https://github.com/Lasagne/Lasagne/blob/5d3c63cb315c50b1cbd27a6bc8664b406f34dd99/docs/conf.py#L113
def linkcode_resolve(domain, info):
    def find_source():
        # try to find the file and line number, based on code from numpy:
        # https://github.com/numpy/numpy/blob/master/doc/source/conf.py#L286
        obj = sys.modules[info["module"]]
        for part in info["fullname"].split("."):
            obj = getattr(obj, part)
        import inspect
        import os

        fn = inspect.getsourcefile(obj)
        fn = os.path.relpath(fn, start=os.path.dirname(multiclass_metrics.__file__))
        source, lineno = inspect.getsourcelines(obj)
        return fn, lineno, lineno + len(source) - 1

    if domain != "py" or not info["module"]:
        return None
    try:
        filename = "multiclass_metrics/%s#L%d-L%d" % find_source()
    except Exception:
        filename = info["module"].replace(".", "/") + ".py"

    # TODO: attempt to read branch from CI config?
    tag = "master"  # if 'dev' in release else ('v' + release)

    return "https://github.com/maximz/multiclass-metrics/blob/%s/%s" % (tag, filename)


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
# see https://alabaster.readthedocs.io/en/latest/customization.html
#
# see also requests docs conf, which uses alabaster theme too: https://github.com/psf/requests/blob/main/docs/conf.py
html_theme_options = {
    ## Alabaster options:
    # "show_powered_by": False,
    # "github_user": "maximz",
    # "github_repo": "multiclass-metrics",
    # "github_banner": True,
    # "github_type": "star",
    # "show_related": False,
    # "note_bg": "#FFF59C",
    ## Furo options:
    "top_of_page_button": None,
    # Since Furo doesn't allow us to disable dark mode, we make dark mode
    # equivalent to light mode by overriding all colors back to their light value.
    # From https://github.com/pradyunsg/furo/issues/28#issuecomment-1178947861
    "dark_css_variables": {
        # Taken from: https://github.com/pradyunsg/furo/blob/c682d5d3502f3fa713c909eebbf9f3afa0f469d9/src/furo/assets/styles/variables/_colors.scss
        "color-problematic": "#b30000",
        # Base Colors
        "color-foreground-primary": "black",  # for main text and headings
        "color-foreground-secondary": "#5a5c63",  # for secondary text
        "color-foreground-muted": "#646776",  # for muted text
        "color-foreground-border": "#878787",  # for content borders
        "color-background-primary": "white",  # for content
        "color-background-secondary": "#f8f9fb",  # for navigation + ToC
        "color-background-hover": "#efeff4ff",  # for navigation-item hover
        "color-background-hover--transparent": "#efeff400",
        "color-background-border": "#eeebee",  # for UI borders
        "color-background-item": "#ccc",  # for "background" items (eg: copybutton)
        # Announcements
        "color-announcement-background": "#000000dd",
        "color-announcement-text": "#eeebee",
        # Brand colors
        "color-brand-primary": "#2962ff",
        "color-brand-content": "#2a5adf",
        # Highlighted text (search)
        "color-highlighted-background": "#ddeeff",
        # GUI Labels
        "color-guilabel-background": "#ddeeff80",
        "color-guilabel-border": "#bedaf580",
        # API documentation
        "color-api-keyword": "var(--color-foreground-secondary)",
        "color-highlight-on-target": "#ffffcc",
        # Admonitions
        "color-admonition-background": "transparent",
        # Cards
        "color-card-border": "var(--color-background-secondary)",
        "color-card-background": "transparent",
        "color-card-marginals-background": "var(--color-background-hover)",
        # Code blocks
        "color-code-foreground": "black",
        "color-code-background": "#f8f9fb",
    },
}
# Force pygments style in dark mode back to the light variant
pygments_dark_style = "tango"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
# html_extra_path = ["tests/baseline"]

# These paths are either relative to html_static_path or fully qualified paths (eg. https://...)
html_css_files = [
    "custom_furo.css",
]

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "multiclass_metricsdoc"


# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "multiclass_metrics.tex",
        "multiclass-metrics Documentation",
        author,
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "multiclass_metrics",
        "multiclass-metrics Documentation",
        [author],
        1,
    )
]

# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "multiclass_metrics",
        "multiclass-metrics Documentation",
        author,
        "multiclass_metrics",
        "Multiclass Metrics",
        "Miscellaneous",
    ),
]



"""Multiclass Metrics."""

from inspect import signature
from typing import Callable, Optional, Tuple, List
import logging
import numpy as np
import pandas as pd
import sklearn.metrics
from sklearn.metrics._base import _average_multiclass_ovo_score, _average_binary_score

__author__ = """Maxim Zaslavsky"""
__email__ = "maxim@maximz.com"
__version__ = "0.0.1"

# Set default logging handler to avoid "No handler found" warnings.
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())

logger = logging.getLogger(__name__)


def _inject_missing_labels(
    y_true: np.ndarray, y_score: np.ndarray, labels: np.ndarray
) -> Tuple[np.ndarray, List[str]]:
    """
    Returns: new probability matrix, class names in order for that matrix

    (for binary classification, converts to and returns probability martix, rather than single vector corresponding to labels[1])

    Arguments:

    - y_true: all ground truth labels
    - y_score: predicted probabilities matrix values
    - labels: column names for y_score (i.e. class names)
    """
    y_score = np.array(y_score)  # defensive cast
    if y_score.ndim == 1 and len(labels) == 2:
        # if y_score has already been converted to a single vector for binary case (for positive class),
        # reshape it to a 2d array with each row being (1-p, p)
        y_score = np.c_[1 - y_score, y_score]

    y_true_classes = set(y_true)
    y_score_classes = set(labels)

    if y_true_classes != y_score_classes:
        y_score = pd.DataFrame(y_score, columns=labels)

        # Are any classes from y_true missing from y_score?
        missing_classes = y_true_classes - y_score_classes
        if len(missing_classes) > 0:
            # Add missing classes to probability matrix
            for c in missing_classes:
                # add each missing class to y_score, as a column of 0s
                logger.warning(f"Inserting phantom class: {c}")
                y_score[c] = 0.0

            # get new labels list, which sklearn's roc_auc_score insists must be in sorted order
            labels = sorted(np.unique(y_score.columns))

            # arrange columns in that sort order
            y_score = y_score[labels]

        # Are any classes from y_score missing from y_true?
        missing_classes_reverse = y_score_classes - y_true_classes
        if len(missing_classes_reverse) > 0:
            # Remove these classes from probability matrix.
            for c in missing_classes_reverse:
                logger.warning(f"Removing class absent from y_true: {c}")
                y_score.drop(c, axis=1, inplace=True)

            # get new labels list, which sklearn's roc_auc_score insists must be in sorted order
            labels = sorted(np.unique(y_score.columns))
            # arrange columns in that sort order
            y_score = y_score[labels]

        # convert back to numpy array
        y_score = np.array(y_score)

    return y_score, labels


def _multiclass_score(
    binary_score_func: Callable,
    y_true: np.ndarray,
    y_score: np.ndarray,
    average: str = "macro",
    labels: Optional[np.ndarray] = None,
    sample_weight: Optional[np.ndarray] = None,
    multi_class: str = "ovo",
) -> float:
    """
    Multiclass probabilistic score with missing labels:

    For multiclass situations where classifier does not learn to predict a certain class, but that class exists in test data: allow running ROC-AUC (or other probabilistic score like auPRC) by inserting missing labels into predict_proba output with probability 0.

    Also handles `y_true` not including all the classes listed in `labels`.

    Falls back to standard score if all labels are present: follows the sklearn implementation but not restricted to ROC AUC and does not require probabilities to sum to 1 (https://github.com/scikit-learn/scikit-learn/blob/36958fb240fbe435673a9e3c52e769f01f36bec0/sklearn/metrics/_ranking.py#L588)

    Parameters:

    - ``y_true``: true multiclass labels as a 1d array.
    - ``y_score``: predicted probabilities for each class as a 2d array. (In binary case, you can instead provide 1d array, like what sklearn metric functions expect.)
    - ``labels``: y_score column names. ``labels`` defaults to ``np.unique(y_true)``.
    - ``average``: can be ``macro`` or ``weighted``.
    - ``multiclass``: can be ``ovo`` or ``ovr``.
    """
    # TODO: Docs for "Also handles `y_true` not including all the classes listed in `labels`"
    y_score = np.array(y_score)  # defensive cast to numpy array

    if labels is None:
        labels = np.unique(y_true)
        # Sanity check:
        if (
            y_score.ndim == 2
            and (y_score.shape[1] > 2 or len(labels) > 2)
            and len(labels) != y_score.shape[1]
        ):
            # Multiclass classification, labels weren't provided, and y_score has a different number of (e.g. more) columns than y_true has unique values.
            raise ValueError(
                "Labels must be provided for this multiclass problem: y_true has fewer uniques than y_score number of columns, so can't infer labels."
            )
    elif len(labels) != len(np.unique(y_true)):
        # For probabilistic scores, y_true's number of unique entries must match the length of labels. That's literally the error text otherwise.
        y_score, labels = _inject_missing_labels(
            y_true=y_true, y_score=y_score, labels=labels
        )

    if len(labels) == 1:
        # Perhaps only one label left after removing phantom classes.
        # Throw same error as roc_auc_score does.
        raise ValueError(
            "Only one class present in y_true. Probability-based score is not defined in that case."
        )

    if (y_score.ndim == 2 and y_score.shape[1] == 2) or y_score.ndim == 1:
        # Binary classification case. len(labels) should be 2 as well.

        if y_score.ndim == 2 and y_score.shape[1] == 2:
            # Cast to 1d array if not already 1d.
            y_score = y_score[:, 1]

        # average_precision_score expects pos_label
        requires_pos_label = "pos_label" in signature(binary_score_func).parameters
        if requires_pos_label and len(labels) != 2:
            # Binary classification (y_score has 2 columns or is 1d),
            # and a choice of positive label is required,
            # but there's a different number of labels (either supplied by user or autogenerated from y_true's unique values).
            # We already checked len(labels) != 1 above. So len(labels) must be > 2.
            raise ValueError(
                "Can't infer positive label for this binary problem: more than 2 labels were provided or were generated from y_true's unique values."
            )
        # Set second class as the positive class.
        kwargs = {"pos_label": labels[1]} if requires_pos_label else {}

        return binary_score_func(y_true, y_score, sample_weight=sample_weight, **kwargs)

    if multi_class == "ovo":
        if sample_weight is not None:
            logger.warning("sample_weight is ignored for OvO multiclass_score")

        # Encode y_true as [0, ..., n_classes - 1] in labels order,
        # because y_true_encoded class "N" values will be matched to column number N of y_score
        y_true_cat = pd.Categorical(y_true, categories=labels)
        y_true_encoded = y_true_cat.codes

        # average_precision_score expects pos_label, but the default pos_label=1 is appropriate,
        # because _average_multiclass_ovo_score's a_true will be 1 when a and 0 when b (all other rows are removed),
        # and vice-versa for b_true, for each pair of classes a and b.
        return _average_multiclass_ovo_score(
            binary_metric=binary_score_func,
            y_true=y_true_encoded,
            y_score=y_score,
            average=average,
        )
    elif multi_class == "ovr":
        return _average_binary_score(
            binary_metric=binary_score_func,
            y_true=sklearn.preprocessing.label_binarize(y_true, classes=labels),
            y_score=y_score,
            average=average,
            sample_weight=sample_weight,
        )
    else:
        raise ValueError("Only OvO and OvR multiclass are supported.")


def roc_auc_score(
    y_true: np.ndarray,
    y_score: np.ndarray,
    average: str = "macro",
    labels: Optional[np.ndarray] = None,
    sample_weight: Optional[np.ndarray] = None,
    multi_class: str = "ovo",
) -> float:
    """
    Area under ROC curve, binary or multiclass OvO.

    Multiclass OvO ROC-AUC is calculated for each pair of classes, then macro or weighted averaged.
    (For each pair of classes, the calculated AUC is the average of the AUCs when each class gets a turn to be the positive class.)

    If binary: falls back to standard AUC as implemented in sklearn.

    Differences from sklearn implementation:

    - Accommodates missing classes in y_true or y_score. (see _multiclass_score docs)
    - Does not require y_score rows to sum to 1.

    Parameters:

    - ``y_true``: true multiclass labels as a 1d array.
    - ``y_score``: predicted probabilities for each class as a 2d array. (In binary case, you can instead provide 1d array, like what sklearn metric functions expect.)
    - ``labels``: y_score columns must be in ``labels`` parameter order. ``labels`` defaults to ``np.unique(y_true)``.
    - ``average``: can be ``macro`` or ``weighted``
    - ``multiclass='ovo'``: only OvO is supported.
    """
    return _multiclass_score(
        binary_score_func=sklearn.metrics.roc_auc_score,
        y_true=y_true,
        y_score=y_score,
        average=average,
        labels=labels,
        multi_class=multi_class,
        sample_weight=sample_weight,
    )


def auprc(
    y_true: np.ndarray,
    y_score: np.ndarray,
    average: str = "macro",
    labels: Optional[np.ndarray] = None,
    sample_weight: Optional[np.ndarray] = None,
    multi_class: str = "ovo",
) -> float:
    """
    Area under precision-recall curve.

    Just like multiclass OvO ROC-AUC, the auPRC is calculated for each pair of classes, then macro or weighted averaged.
    (For each pair of classes, the calculated auPRC is the average of the auPRCs when each class gets a turn to be the positive class.)

    If binary: falls back to standard auPRC as implemented in sklearn.

    Parameters:

    - ``y_true``: true multiclass labels as a 1d array.
    - ``y_score``: predicted probabilities for each class as a 2d array. (In binary case, you can instead provide 1d array, like what sklearn metric functions expect.)
    - ``labels``: y_score columns must be in ``labels`` parameter order. ``labels`` defaults to ``np.unique(y_true)``.
    - ``average``: can be ``macro`` or ``weighted``
    - ``multiclass='ovo'``: only OvO is supported.
    """
    return _multiclass_score(
        binary_score_func=sklearn.metrics.average_precision_score,
        y_true=y_true,
        y_score=y_score,
        average=average,
        labels=labels,
        multi_class=multi_class,
        sample_weight=sample_weight,
    )



import os
import pytest

if os.getenv("_PYTEST_RAISE", "0") != "0":
    # For debugging tests with pytest and vscode:
    # Configure pytest to not swallow exceptions, so that vscode can catch them before the debugging session ends.
    # See https://stackoverflow.com/a/62563106/130164
    # The .vscode/launch.json configuration should be:
    # "configurations": [
    #     {
    #         "name": "Python: Debug Tests",
    #         "type": "python",
    #         "request": "launch",
    #         "program": "${file}",
    #         "purpose": ["debug-test"],
    #         "console": "integratedTerminal",
    #         "justMyCode": false,
    #         "env": {
    #             "_PYTEST_RAISE": "1"
    #         },
    #     },
    # ]
    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value



import numpy as np
import pandas as pd
import pytest
import multiclass_metrics
import sklearn.metrics


def test_roc_auc_score_with_missing_labels_in_yscore():
    # confirm main functionality: handle missing classes automatically
    assert (
        multiclass_metrics.roc_auc_score(
            y_true=["Healthy", "Ebola", "HIV", "Healthy", "Covid"],
            y_score=np.array(
                [
                    [0.1, 0.1, 0.8],
                    [0.33, 0.33, 0.34],
                    [0.1, 0.8, 0.1],
                    [0.05, 0.05, 0.9],
                    [0.8, 0.1, 0.1],
                ]
            ),
            labels=["Covid", "HIV", "Healthy"],
            multi_class="ovo",
            average="macro",
        )
        == 0.875
        # and confirm that adding missing classes as columns of 0s does not change ROC score
        == sklearn.metrics.roc_auc_score(
            y_true=["Healthy", "Ebola", "HIV", "Healthy", "Covid"],
            y_score=np.array(
                [
                    [0.1, 0.0, 0.1, 0.8],
                    [0.33, 0.0, 0.33, 0.34],
                    [0.1, 0.0, 0.8, 0.1],
                    [0.05, 0.0, 0.05, 0.9],
                    [0.8, 0.0, 0.1, 0.1],
                ]
            ),
            labels=["Covid", "Ebola", "HIV", "Healthy"],
            multi_class="ovo",
            average="macro",
        )
        # and confirm that this is different from removing the entries for the missing classes
        != sklearn.metrics.roc_auc_score(
            y_true=["Healthy", "HIV", "Healthy", "Covid"],
            y_score=np.array(
                [
                    [0.1, 0.1, 0.8],
                    [0.1, 0.8, 0.1],
                    [0.05, 0.05, 0.9],
                    [0.8, 0.1, 0.1],
                ]
            ),
            labels=["Covid", "HIV", "Healthy"],
            multi_class="ovo",
            average="macro",
        )
    )

    # pass through if no missing classes
    assert (
        multiclass_metrics.roc_auc_score(
            y_true=["Healthy", "HIV", "Healthy", "Covid"],
            y_score=np.array(
                [
                    [0.1, 0.1, 0.8],
                    [0.1, 0.8, 0.1],
                    [0.05, 0.05, 0.9],
                    [0.8, 0.1, 0.1],
                ]
            ),
            labels=["Covid", "HIV", "Healthy"],
            multi_class="ovo",
            average="macro",
        )
        == sklearn.metrics.roc_auc_score(
            y_true=["Healthy", "HIV", "Healthy", "Covid"],
            y_score=np.array(
                [
                    [0.1, 0.1, 0.8],
                    [0.1, 0.8, 0.1],
                    [0.05, 0.05, 0.9],
                    [0.8, 0.1, 0.1],
                ]
            ),
            labels=["Covid", "HIV", "Healthy"],
            multi_class="ovo",
            average="macro",
        )
        == 1.0
    )

    # further confirmation that adding columns of 0 probabilities (for missing classes) does not change ROC AUC
    y_true = np.array([0, 0, 1, 2])
    y_scores = np.array(
        [[0.1, 0.9, 0.0], [0.3, 0.6, 0.1], [0.35, 0.6, 0.05], [0.8, 0.2, 0.0]]
    )
    y_scores2 = np.array(
        [
            [0.1, 0.9, 0.0, 0.0],
            [0.3, 0.6, 0.1, 0.0],
            [0.35, 0.6, 0.05, 0.0],
            [0.8, 0.2, 0.0, 0.0],
        ]
    )
    assert (
        sklearn.metrics.roc_auc_score(
            y_true, y_scores, multi_class="ovo", labels=[0, 1, 2], average="macro"
        )
        == sklearn.metrics.roc_auc_score(
            y_true, y_scores2, multi_class="ovo", labels=[0, 1, 2, 3], average="macro"
        )
        == 0.25
    )
    assert (
        sklearn.metrics.roc_auc_score(
            y_true, y_scores, multi_class="ovo", labels=[0, 1, 2], average="weighted"
        )
        == sklearn.metrics.roc_auc_score(
            y_true,
            y_scores2,
            multi_class="ovo",
            labels=[0, 1, 2, 3],
            average="weighted",
        )
        == 0.21875
    )


def test_roc_auc_with_missing_labels_must_be_multiclass_ovo_or_ovr_mode():
    y_true = ["Healthy", "Ebola", "HIV", "Healthy", "Covid"]
    labels = ["Covid", "HIV", "Healthy"]
    y_score = np.array(
        [
            [0.1, 0.1, 0.8],
            [0.33, 0.33, 0.34],
            [0.1, 0.8, 0.1],
            [0.05, 0.05, 0.9],
            [0.8, 0.1, 0.1],
        ]
    )
    for multi_class in ["ovr", "ovo"]:
        # no error:
        multiclass_metrics.roc_auc_score(
            y_true=y_true,
            y_score=y_score,
            labels=labels,
            multi_class=multi_class,
        )

    # error:
    with pytest.raises(ValueError, match="Only OvO and OvR multiclass are supported"):
        multiclass_metrics.roc_auc_score(
            y_true=y_true,
            y_score=y_score,
            labels=labels,
            multi_class="something else",
        )


def test_roc_auc_score_doesnt_have_to_sum_to_one():
    assert multiclass_metrics.roc_auc_score(
        y_true=["Covid", "HIV", "Healthy"],
        y_score=np.array(
            [
                [0.1, 0.1, 0.8],
                [0.33, 0.33, 0.34],
                [0.1, 0.8, 0.1],
            ]
        ),
        labels=["Covid", "HIV", "Healthy"],
        multi_class="ovo",
        average="macro",
    ) == multiclass_metrics.roc_auc_score(
        y_true=["Covid", "HIV", "Healthy"],
        y_score=np.array(
            [
                [0.2, 0.1, 0.8],
                [0.66, 0.33, 0.34],
                [0.2, 0.8, 0.1],
            ]
        ),
        labels=["Covid", "HIV", "Healthy"],
        multi_class="ovo",
        average="macro",
    )


def test_roc_auc_score_with_missing_labels_in_ytest():
    # confirm main functionality: handle missing classes automatically
    # also tests that roc auc score does not have to sum to 1
    assert np.allclose(
        [
            multiclass_metrics.roc_auc_score(
                y_true=["Covid", "Covid", "Covid", "Healthy", "Healthy"],
                y_score=np.array(
                    [
                        [0.1, 0.1, 0.8],
                        [0.4, 0.5, 0.1],
                        [0.8, 0.1, 0.1],
                        [0.33, 0.33, 0.34],
                        [0.2, 0.2, 0.6],
                    ]
                ),
                labels=["Covid", "HIV", "Healthy"],
                multi_class="ovo",
                average="macro",
            ),
            # and confirm that removing missing classes does not change ROC score (though sklearn requires renorm to sum to 1)
            sklearn.metrics.roc_auc_score(
                y_true=["Covid", "Covid", "Covid", "Healthy", "Healthy"],
                y_score=np.array(
                    [
                        [0.2, 0.8],
                        [0.8, 0.2],
                        [0.8, 0.2],
                        [0.49, 0.51],
                        [0.25, 0.75],
                    ]
                )[:, 1],
                labels=["Covid", "Healthy"],
                multi_class="ovo",
                average="macro",
            ),
        ],
        2 / 3,
    )


def test_roc_auc_score_with_so_many_missing_labels_that_only_one_label_is_left():
    # 3 initial categorical labels
    # but all y_true belongs to a single label
    y_true = pd.Series(
        ["surviving_class", "surviving_class"],
        dtype=pd.CategoricalDtype(
            categories=["surviving_class", "other_class", "dropped_class"]
        ),
    )
    # classifier only had 2 of the 3 initial classes
    y_score = np.array([[0.6, 0.4], [0.5, 0.5]])
    labels = np.array(["surviving_class", "dropped_class"])

    y_score_modified, labels_modified = multiclass_metrics._inject_missing_labels(
        y_true=y_true,
        y_score=y_score,
        labels=labels,
    )
    assert np.array_equal(y_score_modified, np.array([[0.6], [0.5]]))
    assert np.array_equal(labels_modified, ["surviving_class"])

    for average in ["weighted", "macro"]:
        with pytest.raises(
            ValueError,
            match="Only one class present in y_true. Probability-based score is not defined in that case.",
        ):
            assert (
                multiclass_metrics.roc_auc_score(
                    y_true=y_true,
                    y_score=y_score,
                    labels=labels,
                    average=average,
                )
                == 0.5
            )


def test_auprc():
    ## Generate sample data: "muddied final column" from test_adjust_model_decision_thresholds:

    # first, clear diagonal
    # how many entries per class are clear diagonal
    n_diagonal_clear = 100
    labels = np.array([0, 1, 2, 3])
    # make labels categorical
    labels = np.array(["class" + str(i) for i in labels])

    clear_diagonal_probas = np.vstack(
        [
            np.tile([0.7, 0.1, 0.1, 0.1], n_diagonal_clear).reshape(-1, 4),
            np.tile([0.1, 0.7, 0.1, 0.1], n_diagonal_clear).reshape(-1, 4),
            np.tile([0.1, 0.1, 0.7, 0.1], n_diagonal_clear).reshape(-1, 4),
            np.tile([0.1, 0.1, 0.1, 0.7], n_diagonal_clear).reshape(-1, 4),
        ]
    )
    clear_diagonal_trues = np.hstack(
        [np.tile([lbl], n_diagonal_clear) for lbl in labels]
    )

    # now, muddy up the final column:
    # all predictions consistent, but ground truth is a toss up
    n_muddy = 100
    muddy_final_row_probas = np.tile([0.1, 0.1, 0.1, 0.7], n_muddy * 4).reshape(-1, 4)
    muddy_final_row_trues = np.hstack([np.tile([lbl], n_muddy) for lbl in labels])

    y_score = np.vstack([clear_diagonal_probas, muddy_final_row_probas])
    y_true = np.hstack([clear_diagonal_trues, muddy_final_row_trues])

    # Sanity check accuracy score
    y_score_argmax = labels[y_score.argmax(axis=1)]
    assert sklearn.metrics.accuracy_score(y_true, y_score_argmax) == 0.625

    # make sure same result even if labels come in different order
    labels_reordered = np.array(["class3", "class0", "class1", "class2"])
    y_score_reordered = pd.DataFrame(y_score, columns=labels)[labels_reordered].values

    assert np.allclose(
        [
            multiclass_metrics.auprc(y_true, y_score),
            multiclass_metrics.auprc(y_true, y_score, labels=labels),
            multiclass_metrics.auprc(
                y_true, y_score_reordered, labels=labels_reordered
            ),
        ],
        0.72916,
    )

    assert not np.allclose(
        [
            multiclass_metrics.auprc(
                y_true, y_score_reordered
            ),  # no label order provided, so the assumed one will be wrong
            multiclass_metrics.auprc(
                y_true, y_score, labels=labels_reordered
            ),  # wrong label order provided
            multiclass_metrics.auprc(
                y_true, y_score_reordered, labels=labels
            ),  # wrong label order provided
        ],
        0.72916,
    )

    # make sure it supports binary too, whether y_score is provided as 2d or 1d array (sklearn style)
    chosen_labels = labels[:2]
    y_true_subselect = pd.Series(y_true)
    subselect_mask = y_true_subselect.isin(chosen_labels)
    y_true_subselect = y_true_subselect[subselect_mask].values
    y_score_subselect = pd.DataFrame(y_score, columns=labels)
    y_score_subselect = y_score_subselect[chosen_labels].loc[subselect_mask].values

    assert np.allclose(
        [
            multiclass_metrics.auprc(y_true_subselect, y_score_subselect),
            multiclass_metrics.auprc(
                y_true_subselect, y_score_subselect, labels=chosen_labels
            ),
            multiclass_metrics.auprc(y_true_subselect, y_score_subselect[:, 1]),
            multiclass_metrics.auprc(
                y_true_subselect, y_score_subselect[:, 1], labels=chosen_labels
            ),
            sklearn.metrics.average_precision_score(
                y_true_subselect, y_score_subselect[:, 1], pos_label=chosen_labels[1]
            ),
            sklearn.metrics.average_precision_score(
                y_true_subselect, y_score_subselect[:, 0], pos_label=chosen_labels[0]
            ),
        ],
        0.75,
    )



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("CHANGELOG.md") as history_file:
    history = history_file.read()

requirements = ["numpy", "pandas", "scikit-learn"]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Maxim Zaslavsky",
    author_email="maxim@maximz.com",
    name="multiclass_metrics",
    description="Multiclass Metrics",
    packages=find_packages(include=["multiclass_metrics", "multiclass_metrics.*"]),
    python_requires=">=3.8",
    version="0.0.1",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    include_package_data=True,
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/maximz/multiclass-metrics",
    zip_safe=False,
)
